﻿using System;
using System.Collections.Generic;
using System.Linq;
using Armory.Settings;
using Zeta.Game;
using Zeta.Game.Internals.Actors;

namespace Armory
{
    internal class ItemEvaluator
    {
        private static HashSet<int> dequippedItems = new HashSet<int>();
        private static DateTime lastEvaluatedItems = DateTime.MinValue;
        private static HashSet<Tuple<InventorySlot, int>> checkedItems = new HashSet<Tuple<InventorySlot, int>>();

        private const int checkIntervalMillseconds = 1000;

        private static int lastBackpackCount = -1;
        private static int lastStashCount = -1;

        private static HashSet<InventorySlot> HirelingSlots = new HashSet<InventorySlot>()
        {
            InventorySlot.HirelingLeftFinger,
            InventorySlot.HirelingLeftHand,
            InventorySlot.HirelingNeck,
            InventorySlot.HirelingRightFinger,
            InventorySlot.HirelingRightHand,
            InventorySlot.HirelingSpecial
        };

        /// <summary>
        /// Iterates through 
        /// </summary>
        public static void EvaluateItems()
        {
            if (!Player.IsValid)
                return;

            if (Armory.ActiveHero == null)
                return;

            //Only pulse out of combat
            if (Zeta.Bot.CombatTargeting.Instance.FirstObject != null)
                return;

            if (DateTime.UtcNow.Subtract(lastEvaluatedItems).TotalMilliseconds < checkIntervalMillseconds && !GameUI.ElementIsVisible(GameUI.StashTab1))
                return;

            lastEvaluatedItems = DateTime.UtcNow;

            // Stash isn't open
            if (!GameUI.ElementIsVisible(GameUI.StashTab1))
            {
                int backpackCount = ZetaDia.Me.Inventory.Backpack.Count();
                if (lastBackpackCount == backpackCount)
                {
                    return;
                }
                else
                    lastBackpackCount = backpackCount;
            }
            // Stash is open
            else
            {
                int stashCount = ZetaDia.Me.Inventory.StashItems.Count();
                if (lastStashCount == stashCount)
                {
                    return;
                }
                else
                    lastStashCount = stashCount;
            }

            // counts are different for backpack or stash, continue

            Player.Update();

            // Iterate through every possible inventory slot
            foreach (InventorySlot slot in Enum.GetValues(typeof(InventorySlot)).Cast<InventorySlot>())
            {
                if (HirelingSlots.Contains(slot))
                    continue;

                if (slot == InventorySlot.None)
                    continue;

                ACDItem item;
                if (!EvaluateItemSlot(slot, out item))
                    continue;

                if (item == null || !item.IsValid)
                    continue;

                EquipItem(slot, item);
            }
        }

        private static void EquipItem(InventorySlot slot, ACDItem item)
        {
            dequippedItems.Add(item.ACDGuid);
            Logger.Log("Equipping item {0} in slot {1}", item.Name, slot);
            ZetaDia.Me.Inventory.EquipItem(item.DynamicId, slot);
            Player.Update();
        }

        public static bool EvaluateItemSlot(InventorySlot slot, out ACDItem foundItem)
        {
            foundItem = default(ACDItem);

            IEnumerable<ACDItem> itemsList;
            if (Settings.CheckStash && GameUI.ElementIsVisible(GameUI.StashTab1))
            {
                itemsList = ZetaDia.Me.Inventory.StashItems;
            }
            else
            {
                itemsList = ZetaDia.Me.Inventory.Backpack;
            }

            itemsList = (from i in itemsList
                         where
                         !checkedItems.Contains(new Tuple<InventorySlot, int>(slot, i.ACDGuid)) &&
                         !dequippedItems.Contains(i.ACDGuid) &&
                         !i.IsUnidentified &&
                         i.ValidInventorySlots.Contains(slot)
                         select i).ToList();

            foreach (ACDItem item in itemsList)
            {
                if (ShouldEquipItem(slot, item))
                {
                    foundItem = item;
                    return true;
                }
            }
            return false;
        }

        public static bool ShouldEquipItem(InventorySlot slot, ACDItem acdItem)
        {
            ItemWrapper item = new ItemWrapper(acdItem);

            checkedItems.Add(new Tuple<InventorySlot, int>(slot, item.ACDGuid));

            // Two handed weapons can never go into RightHand
            if (slot == InventorySlot.RightHand && item.IsTwoHand)
                return false;

            if (!Settings.AllowTwoHand && item.IsTwoHand)
                return false;

            if (!Settings.AllowShields && item.IsShield)
                return false;

            if (!Settings.DualWield && item.IsOneHand && slot == InventorySlot.RightHand)
                return false;

            if (Settings.DisableAt70 && Player.Level >= 70)
                return false;

            if (!IsSlotProtected(slot, item))
                return false;

            // Is an equippable item type for our Hero?
            if (!Armory.ActiveHero.CanEquipItem(item, slot))
                return false;

            // We meet the level requirements?
            if (acdItem.Level > ZetaDia.Me.Level)
            {
                Logger.Debug("We don't meet the level requirements for {0} ({1})", item.Name, acdItem.Level);
                return false;
            }

            // Check if our slot is empty
            bool isSlotEmpty = !ZetaDia.Me.Inventory.Equipped.Any(i => i.InventorySlot == slot);

            bool isUpgrade = false;

            float damage = 0f, toughness = 0f, healing = 0f;
            Logger.Debug("Evaluating {0} - {1}", slot, item.Name);

            if (slot == InventorySlot.RightFinger)
            {
                // Alt item slot for Right Ring finger
                acdItem.GetStatChanges(out damage, out healing, out toughness, true);
            }
            else if (slot == InventorySlot.RightHand && !Player.HasTwoHandEquipped && (Player.IsDualWielding || isSlotEmpty))
            {
                // Alt Item Slot for Right Hand if dual wielding or slot is empty and no 2H equipped
                acdItem.GetStatChanges(out damage, out healing, out toughness, true);
            }
            else
            {
                acdItem.GetStatChanges(out damage, out healing, out toughness, false);
            }

            isUpgrade = CalculateUpgrade(damage, toughness, healing);

            if (isSlotEmpty)
            {
                Logger.Log("{0} is an Upgrade -  Slot is Empty", item.Name);
                return true;
            }

            if (isUpgrade)
            {
                Logger.Log("{0} is an Upgrade! Damage={1:0.0}% Toughness={2:0.0}% Healing={3:0.0}%", item.Name, damage * 100, toughness * 100, healing * 100);
            }
            return isUpgrade;
        }

        private static bool CalculateUpgrade(float damage, float toughness, float healing)
        {
            bool isUpgrade = false;

            // Additive Damage Formula - Damage reduced up to 5% but toughness compensates
            if (Settings.UseAdditiveDamageFormula &&
                damage > Settings.AdditiveDamageFormulaDamage &&
                damage + toughness > Settings.AdditiveDamageFormulaDamageToughness)
                isUpgrade = true;

            // Healing Formula - Damage reduced max 15% but toughness buff is "good enough" and healing is > 0%
            if (Settings.UseHealingUpgradeFormula &&
                damage > Settings.HealingUpgradeFormulaDamage &&
                damage + toughness > Settings.HealingUpgradeFormulaDamageToughness &&
                healing > Settings.HealingUpgradeFormulaHealing)
                isUpgrade = true;

            // Basic Upgrade Formula
            if (Settings.UseBasicUpgradeFormula &&
                damage > Settings.BasicUpgradeFormulaDamage &&
                toughness > Settings.BasicUpgradeFormulaToughness)
                isUpgrade = true;

            /*
             *  Want to DIY? Uncomment below as needed!
             */
            // Additive Damage Formula - Damage reduced up to 5% but toughness compensates
            //if (damage > -0.05f && damage + toughness > 0)
            //    isUpgrade = true;

            // Healing Formula - Damage reduced max 15% but toughness buff is "good enough" and healing is > 0%
            //if (damage > -0.15f && damage + toughness > 0 && healing > 0)
            //    isUpgrade = true;

            // Basic Upgrade Formula
            //if (damage > 0 && toughness > 0)
            //    isUpgrade = true;

            return isUpgrade;
        }

        private static ArmorySettings Settings
        {
            get { return ArmorySettings.Instance; }
        }

        private static bool IsSlotProtected(InventorySlot slot, ItemWrapper item)
        {
            if (Settings.ProtectedSlots.Shoulders && item.InventorySlot == InventorySlot.Shoulders)
                return false;

            if (Settings.ProtectedSlots.Helm && item.InventorySlot == InventorySlot.Head)
                return false;

            if (Settings.ProtectedSlots.Neck && item.InventorySlot == InventorySlot.Neck)
                return false;

            if (Settings.ProtectedSlots.Hands && item.InventorySlot == InventorySlot.Hands)
                return false;

            if (Settings.ProtectedSlots.Chest && item.InventorySlot == InventorySlot.Torso)
                return false;

            if (Settings.ProtectedSlots.Wrists && item.InventorySlot == InventorySlot.Bracers)
                return false;

            if (Settings.ProtectedSlots.Ring1 && item.InventorySlot == InventorySlot.LeftFinger)
                return false;

            if (Settings.ProtectedSlots.Waist && item.InventorySlot == InventorySlot.Waist)
                return false;

            if (Settings.ProtectedSlots.Ring2 && item.InventorySlot == InventorySlot.RightFinger)
                return false;

            if (Settings.ProtectedSlots.Legs && item.InventorySlot == InventorySlot.Legs)
                return false;

            if (Settings.ProtectedSlots.Feet && item.InventorySlot == InventorySlot.Feet)
                return false;

            return true;
        }

        #region GetGearvalue
        [Obsolete("GetGearValue is not updated for RoS")]
        public static double GetGearValue(double DPS, double EHP, double LS)
        {
            // Calculation rewriting to be done on below code when DPS/EHP are accessible here (rethink: damagefactor/ehpfactor, perhaps go for preferred EHP value?)
            double val = 0;
            double damageFactor = 50;
            if (ZetaDia.Me.Level != 70)
            {
                switch (ZetaDia.Me.ActorClass)
                {
                    case ActorClass.Crusader:
                    case ActorClass.Barbarian:
                    case ActorClass.Monk:
                        damageFactor *= 0.8;
                        break;
                    case ActorClass.DemonHunter:
                    case ActorClass.Wizard:
                    case ActorClass.Witchdoctor:
                        break;
                }
                val = ((DPS * 50) * (EHP / (DPS * damageFactor))) + (EHP * ((DPS * damageFactor) / EHP));
            }
            if (ZetaDia.Me.Level == 70)
            {
                if (LS < 1)
                {
                    LS = 1;
                }
                double EHPscaler = Math.Sqrt((EHP - 300000) / 100000);
                if (EHP < 400000)
                {
                    EHPscaler = Math.Sqrt(EHP / 400000);
                }
                double LSvalue = Math.Sqrt(LS) * (DPS / 150000);
                if (LSvalue < 1)
                {
                    LSvalue = 1;
                }
                val = (((DPS * EHPscaler) * damageFactor) + (EHP / EHPscaler) * LSvalue);
            }
            return val;
        }
        #endregion

        public static void Reset()
        {
            dequippedItems.Clear();
            lastEvaluatedItems = DateTime.MinValue;
            checkedItems.Clear();
        }

        public static readonly List<ItemType> GenericEquippableItemTypes = new List<ItemType>()
        {
            ItemType.Amulet,
            ItemType.Belt,
            ItemType.Boots,
            ItemType.Bracer,
            ItemType.Chest,
            ItemType.Gloves,
            ItemType.Helm,
            ItemType.Legs,
            ItemType.Ring,
            ItemType.Shield,
            ItemType.Shoulder,
        };

        public static readonly List<ItemType> GenericClassItemTypes = new List<ItemType>()
        {
            ItemType.Axe,
            ItemType.Bow,
            ItemType.CeremonialDagger,
            ItemType.Cloak,
            ItemType.Crossbow,
            ItemType.Dagger,
            ItemType.Daibo,
            ItemType.FistWeapon,
            ItemType.HandCrossbow,
            ItemType.Mace,
            ItemType.MightyBelt,
            ItemType.MightyWeapon,
            ItemType.Mojo,
            ItemType.Orb,
            ItemType.Polearm,
            ItemType.Quiver,
            ItemType.Spear,
            ItemType.SpiritStone,
            ItemType.Staff,
            ItemType.Sword,
            ItemType.VoodooMask,
            ItemType.Wand,
            ItemType.WizardHat,
        };

        public static readonly List<ItemType> GenericUnEquippableItemTypes = new List<ItemType>()
        {
            ItemType.CraftingPage,
            ItemType.CraftingPlan,
            ItemType.CraftingReagent,
            ItemType.FollowerSpecial,
            ItemType.Gem,
            ItemType.Potion,
            ItemType.Unknown,
        };
    }
}